///////////////////////////////////////
//  copyright Alexandre Chouvellon 2009
//  Source libre de droit
///////////////////////////////////////
#include <stdlib.h>
#include <stdio.h>
#include "list.h"
#include "structure.h"
#include "language.fwd.h"

static void sendMemoryError()
{
	printf(glb_outOfMemory);
	exit(0);
}

void freeFifo(MyFifoList * fifo)
{
	if(fifo && fifo->start)
	{
		ChainedList * current = fifo->start;
		ChainedList * tobefreed= NULL;
		while(current)
		{
			tobefreed = current;
			current = current->next;
			free(tobefreed);
		}
		free(fifo);
	}
}
static MyFifoList * allocFifo()
{
	MyFifoList * fifo = calloc(1,sizeof(MyFifoList));

	if (!fifo)
		sendMemoryError();

	return fifo;
}

static ChainedList * allocChainedList()
{
	ChainedList *  list = NULL; 

	list = (ChainedList * )calloc(1,sizeof(ChainedList));

	if (!list)
		sendMemoryError();

	return  list;
}
Element * allocElement()
{
	Element  *  elem = NULL; 

	elem = (Element *)calloc(1,sizeof(struct Element));

	if (!elem)
		sendMemoryError();

	return  elem;
}

//pour mettre une lment  la fin de la fifo
void pushToEnd(MyFifoList ** pfifo, Element *  elem)
{

	if (elem && pfifo)
	{
		MyFifoList * fifo = *pfifo;
		ChainedList * cl = allocChainedList();

		cl->element = elem;
		if (!fifo)   {fifo = *pfifo = allocFifo();}
		if (!fifo->start) { fifo->start = cl; }
		if ( fifo->end  )  fifo->end->next = cl; 

		fifo->end = cl;
	}
}


//pop standard d"un fifo 
ChainedList * pop(MyFifoList * fifo)
{
	ChainedList * elem = fifo->start?fifo->start:NULL;

	if (elem)
	{
		fifo->start = elem->next;

		if ( fifo->end == elem)
			fifo->end = NULL;
	}
	return elem;
}

static ChainedList * newChainedElement(Board * board, long long * key1, short key2, Action action)
{
	ChainedList * cl = allocChainedList();
	cl->element = allocElement();
	copyBoard(board, &(cl->element->board));
	cl->element->key1 = *key1;
	cl->element->key2 = key2 ;
	if(action != -1){cl->element->actionDone = action;}
	cl->next = NULL;

	return cl;
}

long long convertRowToType(short int row)
{
	long long res = 0;

	res |= (0x000000000FFFF & getTokenTypeFromId((row & MASK0) >> OFFSET0)) << OFFSET0;
	res |= (0x000000000FFFF & getTokenTypeFromId((row & MASK1) >> OFFSET1)) << OFFSET1;
	res |= (0x000000000FFFF & getTokenTypeFromId((row & MASK2) >> OFFSET2)) << OFFSET2;
	res |= (0x000000000FFFF & getTokenTypeFromId((row & MASK3) >> OFFSET3)) << OFFSET3;
	return res;
}

long long buildPrimaryKey(Board * board)
{
	long long key1  =0;
	int i =0;
	short int row = 0;
	for (i =3; i >=0; --i)
	{

		switch(i)
		{
		case 0:row =  board->row0;break;
		case 1:row =  board->row1;break;
		case 2:row =  board->row2;break;
		case 3:row =  board->row3;break;
		}
		key1 |= (convertRowToType(row)) << 16*i;
	}

	return key1;
}

void freeChainedList(ChainedList * list)
{
	if (list)
	{ 
		ChainedList * current = list;
		ChainedList * tobefreed= NULL;
		while(current)
		{

			tobefreed = current;
			free(current->element);
			current = current->next;
			free(tobefreed);
		}
	}
}
//push tri par ordre croissant key1 puis key2 
//renvoie l'lment trouv e tindique si il a t ajout
ChainedList * pushMe(ChainedList * head, Board * board, Action action,  struct Element ** findElement, char * added)
{
	long long key1 = buildPrimaryKey(board);
	short int key2 = (short int)convertRowToType(board->row4);

	ChainedList * current = head;

	if (added)
		*added = FALSE;
	if(findElement)
		*findElement = NULL;

	if (head == NULL) //cas de base il faut initialiser la liste avec un lment
	{
		current = newChainedElement(board, &key1, key2, action);

		if (added ) *added = TRUE; 
		if (findElement) *findElement = current->element;

		return current;
	}

	//il y a deja une liste de construite, il faut insrer notre element au bon endroit 
	while (current)
	{
		if (current->element->key1 > key1 || (current->element->key1 == key1 && current->element->key2 > key2)) //notre element doit etre insr
		{
			ChainedList * newCl = newChainedElement(board, &key1, key2, action);
			//protg dans l'alloc
			struct Element * element = newCl->element; // l'adresse du nouveau element, on va l'utiliser pour stocker nos nouvelles donnes

			newCl->element = current->element;
			current->element = element;
			newCl->next = current->next;
			current->next = newCl;

			if (added) *added = TRUE;
			if (findElement) *findElement = element;

			return head;
		}

		if ((current->element->key1 == key1 && current->element->key2 == key2)) //l'element courant est dj dans la liste
		{
			if (added) *added = FALSE;
			if (findElement) *findElement = current->element;
			return head;
		}

		if (current->next) //on passe  l'lment suivant si existant
		{
			current = current->next;
		}
		else{
			//la liste touche  sa fin il faut insrer
			ChainedList * cl = newChainedElement(board, &key1, key2, action);

			current->next = cl;
			if (added) *added = TRUE;
			if (findElement) *findElement =  cl->element;
			return head;
		}
	}
	//jamais atteint
	return head;
}


Element * getPreviousElement(Element * elem)
{
	return elem?elem->previousState:NULL;
}

Action getActionDone(Element * elem)
{
	return elem?elem->actionDone:-1;
}

void setPreviousElement(Element * elem, Element * d)
{
	elem->previousState=d;
}

void setActionDone(Element * elem, Action a )
{
	elem->actionDone =a;;
}
Board * getBoard(Element * elem)
{
	return elem?&(elem->board):NULL;
}


Location * getBlank1(Element * e)
{
	return e?&(e->blank1):NULL;
}

Location * getBlank2(Element * e)
{
	return e?&(e->blank2):NULL;
}

Element * getElement(ChainedList * a)
{
	return a?a->element:NULL;
}
