///////////////////////////////////////
//  copyright Alexandre Chouvellon 2009
//  Source libre de droit
///////////////////////////////////////

#include <stdio.h>
#include "tools.h"
#include "donkey.h"
#include "analysis.h"
#include "structure.fwd.h" 
#include "list.fwd.h"
#include "language.fwd.h"
#include "errorId.h"
extern Token glb_tokens[12];

int checkIsAZLetter(char a)
{
   return a >= 'A' || a<= 'L';
}

//complete le type des doubles  (met en vertical ou horizontal) vrifie aussi que le quad n'est pas clat
int checkAndSetToken(Board *board,short int  id, char i, char j)
{
   int ok = TRUE;
   switch(id)
   {
      //les blank
   case 0:
   case 1: break;
      //les simples
   case 2:
   case 3:
   case 4:
   case 5: break;
      //les doubles, cas non trivial il faut regarder si une case a droite ou en bas a le mme id pour diffrencier l'horizontal du vertical
   case 6:
   case 7:
   case 8:
   case 9: 
   case 10:if  (j+1 < 4 && id == getIdfromBoard(board, i , j +1)) {setType(id, DOUBLE_HORIZONTAL);return TRUE;}
		   if  (i+1 < 5 && id == getIdfromBoard(board, i +1 , j)) {setType(id, DOUBLE_VERTICAL); return TRUE;}  printf(glb_errorDefToken);return FALSE;
      //quad
   case 11:  if(id == getIdfromBoard(board, i , j +1) &&  id == getIdfromBoard(board, i +1 , j)&&  id == getIdfromBoard(board, i +1 , j+1)) return TRUE; else {printf(glb_errorDefToken); return FALSE;}
   default: ok = FALSE;
   }

   return ok ;
}

int  loadFile(FILE * file, Board * board)
{
   int ok = TRUE;
   char a0=0, a1=0, a2=0,a3=0;
   int i =0, count;

   for (; i < MAX_ROW; ++i)
   {
      count = fscanf_s(file, "%c%c%c%c\r\n",&a0,sizeof(char),&a1,sizeof(char),&a2,sizeof(char),&a3,sizeof(char));
      if (count!=4)
      {
         printf(glb_atleastExpected);
         return FALSE;
      }
      ok = checkIsAZLetter(a0) && checkIsAZLetter(a1) && checkIsAZLetter(a2) && checkIsAZLetter(a3);
      if( ok )
      {
         short int row = (a0 -'A') << OFFSET0 | (a1 -'A') << OFFSET1 | (a2 -'A') << OFFSET2 | (a3 -'A');

         setRow(board,row,i);
      }
      else{
         printf(glb_badInputLetters);
         break;
      }
   }
   if(EOF!=fscanf_s(file,"%c",&a0))
   {
      printf(glb_extraLines);
      return FALSE;
   }
	fclose(file);
   return ok;
}


int checkConsistancy(Board *board)
{
	int count[12]= {0};
	int check =0;
	char i,j;
	short int id;
	int validate = TRUE;


	count[(getRow(board, 0) & MASK0) >> OFFSET0]++;
	count[(getRow(board, 0) & MASK1) >> OFFSET1]++;
	count[(getRow(board, 0) & MASK2) >> OFFSET2]++;
	count[(getRow(board, 0) & MASK3) >> OFFSET3]++;

	count[(getRow(board, 1) & MASK0) >> OFFSET0]++;
	count[(getRow(board, 1) & MASK1) >> OFFSET1]++;
	count[(getRow(board, 1) & MASK2) >> OFFSET2]++;
	count[(getRow(board, 1) & MASK3) >> OFFSET3]++;

	count[(getRow(board, 2) & MASK0) >> OFFSET0]++;
	count[(getRow(board, 2) & MASK1) >> OFFSET1]++;
	count[(getRow(board, 2) & MASK2) >> OFFSET2]++;
	count[(getRow(board, 2) & MASK3) >> OFFSET3]++;

	count[(getRow(board, 3) & MASK0) >> OFFSET0]++;
	count[(getRow(board, 3) & MASK1) >> OFFSET1]++;
	count[(getRow(board, 3) & MASK2) >> OFFSET2]++;
	count[(getRow(board, 3) & MASK3) >> OFFSET3]++;

	count[(getRow(board, 4) & MASK0) >> OFFSET0]++;
	count[(getRow(board, 4) & MASK1) >> OFFSET1]++;
	count[(getRow(board, 4) & MASK2) >> OFFSET2]++;
	count[(getRow(board, 4) & MASK3) >> OFFSET3]++;

	check =  count[BLANK1] == 1 && count[BLANK2] == 1 &&  count[2] == 1 && count[3] == 1 && count[4] == 1 && count[5] == 1 && count[6] == 2 && count[7] == 2 && count[8] == 2 &&count[9] == 2 && count[10] == 2 && count[11] == 4;

	if(!check)
	{
		printf(glb_badLettersCount);
		return FALSE;
	}

	//on va vrifier que les pieces ne sont pas casses
	//on parcourt de gauche a droite de haut en bas, des qu'on rencontre un nouvel id on l'tudie au cas par cas

	for(i=0; i < 12;++i)
	{
		count[i]=0;
	}
	for(i=0; i < 5 && validate; ++i)
	{
		for(j=0; j < 4 && validate; ++j)
		{
			id = getIdfromBoard(board, i , j);

			if(!count[id]) //pas encore tudi
			{
				setLocation(getLocationToken(id), i, j);
				setType(id, getTypeFromId(id));
				count[id] = 1;
				validate = checkAndSetToken(board, id, i,j);
			}
			else
			{
				count[id]=1;
   			}
		}
	}
	return validate;
}

static char * getReadableOrientation(short orientation)
{
   switch(orientation)
   {
   case TOP:		return glb_top;
   case RIGHT:		return glb_right;
   case BOTTOM:		return glb_bottom;
   case LEFT:		return glb_left;
   }
   printf(glb_errorMessage,ERROR_READABLE_ORIENTATION);
   return "";
}

static void printRow(short int r)
{
   printf(" %c %c %c %c \n", getLetterFromId((r & MASK0) >>OFFSET0), getLetterFromId((r & MASK1) >>OFFSET1), getLetterFromId((r & MASK2) >>OFFSET2), getLetterFromId((r & MASK3) >>OFFSET3));
}

static void printAction(Action a)
{
   if (a != -1)
   {
      short int orientation = (a & MASK2) >> 4; 
      short int id = (a & MASK3); 

      printf(glb_move, getLetterFromId(id), getReadableOrientation(orientation));
   }
}
static void printBoard(Board *b)
{ 
	int i;
	for(i=0; i < MAX_ROW;++i) 
		printRow(getRow(b, i));
}

int explainSolution(Element *elem, int deep)
{
   Element * previousState = getPreviousElement(elem);
   int count = previousState?explainSolution(previousState, 1+deep):deep;
   

   switch(deep)
   {
   case 0:    printf(glb_lastMove);break;
   case 1:    printf(glb_beforeLast);break;
   default :  printf(glb_moveCount, deep +1);break;
   }

   printAction(getActionDone(elem));
   printBoard(getBoard(elem));

   return count;
}
